/*
 * Created by SharpDevelop.
 * User: grunwald
 * Date: 24.07.2006
 * Time: 15:12
 * 
 * To change this template use Tools | Options | Coding | Edit Standard Headers.
 */
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Globalization;
using System.IO;
using System.Text;
using PPJ.Runtime.Scripting;

namespace ScaeParser
{
	internal sealed class ScaeLexer
	{
		readonly TextReader reader;
		int line = 1, column = 0;
		Token nextToken;

		public ScaeLexer(TextReader codeReader)
		{
			this.reader = codeReader;
		}

		public ScaeLexer(string code)
		{
			this.reader = new StringReader(code);
		}

		public Token ReadNextToken()
		{
			if (nextToken == null)
			{
				return ReadToken();
			}
			else
			{
				Token t = nextToken;
				nextToken = null;
				return t;
			}
		}

		/// <summary>
		/// Gets the look-ahead token
		/// </summary>
		public Token NextToken
		{
			[DebuggerStepThrough]
			get
			{
				if (nextToken == null)
				{
					nextToken = ReadToken();
				}
				return nextToken;
			}
		}

		public Location Location
		{
			get
			{
				return new Location(line, column);
			}
		}

		char Peek()
		{
			int p = reader.Peek();
			if (p != -1)
				return (char)p;
			else
				return '\0';
		}

		char Read()
		{
			int r = reader.Read();
			if (r == '\n')
			{
				line += 1;
				column = 0;
				return '\n';
			}
			else if (r != -1)
			{
				column++;
				return (char)r;
			}
			else
			{
				return '\0';
			}
		}

		internal ParserException Error(string message, params object[] args)
		{
			return new ParserException(string.Format(message, args), line, column);
		}

		Token ReadToken()
		{
			char c = Read();
			while (c == '\r' || c == '\t' || c == ' ')
			{
				// ignore whitespace
				c = Read();
			}
			switch (c)
			{
				case '\0':
					return new Token(TokenKind.Eof);
				case '\n':
					return new Token(TokenKind.Newline);
				case '+':
					return new Token(TokenKind.Plus);
				case '-':
					return new Token(TokenKind.Minus);
				case '*':
					return new Token(TokenKind.Multiplication);
				case '/':
					return new Token(TokenKind.Division);
				case ',':
					return new Token(TokenKind.Comma);
				case '(':
					return new Token(TokenKind.OpenParenthesis);
				case ')':
					return new Token(TokenKind.CloseParenthesis);
				case '[':
					return new Token(TokenKind.OpenBracket);
				case ']':
					return new Token(TokenKind.CloseBracket);
				case '=':
					return new Token(TokenKind.Equals);
				case '&':
					return new Token(TokenKind.BitAnd);
				case '|':
					if (Peek() == '|')
					{
						Read();
						return new Token(TokenKind.StringConcat);
					}
					else
					{
						return new Token(TokenKind.BitOr);
					}
				case '!':
					c = Read();
					if (c == '=')
					{
						return new Token(TokenKind.Inequals);
					}
					else
					{
						// ! Kommentar : lese bis Ende der Zeile
						while (c != '\n' && c != '\0')
						{
							c = Read();
						}
						return ReadToken();
					}
				case '>':
					if (Peek() == '=')
					{
						Read();
						return new Token(TokenKind.GreaterThanOrEqual);
					}
					else
					{
						return new Token(TokenKind.GreaterThan);
					}
				case '<':
					if (Peek() == '=')
					{
						Read();
						return new Token(TokenKind.LessThanOrEqual);
					}
					else
					{
						return new Token(TokenKind.LessThan);
					}
				case '"':
				case '\'':
					return ReadStringLiteral(c);
				case '.':
					c = Peek();
					if (c >= '0' && c <= '9')
					{
						return ReadNumber('.');
					}
					else
					{
						return new Token(TokenKind.Dot);
					}
				default:
					if (c == '_' || char.IsLetter(c))
					{
						return ReadIdentifier(c);
					}
					else if (c >= '0' && c <= '9')
					{
						return ReadNumber(c);
					}
					else
					{
						throw Error(Errors.UnknownCharacter, c);
					}
			}
		}

		const int MaxIdentLength = 512;
		char[] identBuffer = new char[MaxIdentLength];

		Token ReadIdentifier(char firstChar)
		{
			identBuffer[0] = firstChar;
			int identLength = 1;
			while (identLength < MaxIdentLength)
			{
				char c = Peek();
				if (char.IsLetterOrDigit(c) || c == '_')
				{
					Read();
					identBuffer[identLength++] = c;
				}
				else
				{
					string name = new string(identBuffer, 0, identLength);
					if (name == "TRUE")
					{
						return new Token(SalNumberLit.One);
					}
					else if (name == "FALSE")
					{
						return new Token(SalNumberLit.Zero);
					}
					else if (string.Equals(name, "and", StringComparison.InvariantCultureIgnoreCase))
					{
						return new Token(TokenKind.And);
					}
					else if (string.Equals(name, "or", StringComparison.InvariantCultureIgnoreCase))
					{
						return new Token(TokenKind.Or);
					}
					else if (string.Equals(name, "not", StringComparison.InvariantCultureIgnoreCase))
					{
						return new Token(TokenKind.Not);
					}
					else
					{
						return new Token(name);
					}
				}
			}
			throw Error(Errors.IdentifierTooLong);
		}

		Token ReadStringLiteral(char quote)
		{
			char c;
			StringBuilder b = new StringBuilder();
			while ((c = Read()) != quote)
			{
				if (c == '\0')
				{
					throw Error(Errors.EndOfFileInsideStringLiteral);
				}
				else if (c == '\\')
				{
					b.Append(Read());
				}
				else
				{
					b.Append(c);
				}
			}
			return new Token(new StringLit(b.ToString(), this.Location));
		}

		Token ReadNumber(char firstDigit)
		{
			bool integer = true;
			StringBuilder b = new StringBuilder();
			char c = firstDigit;
			if (c >= '0' && c <= '9')
			{
				b.Append(c);
				c = Peek();
				while (c >= '0' && c <= '9')
				{
					b.Append(c);
					Read();
					c = Peek();
				}
			}
			if (c == '.')
			{
				b.Append('.');
				integer = false;
				if (firstDigit != '.')
				{
					Read();
				}
				c = Peek();
				if (c < '0' || c > '9')
				{
					throw Error(Errors.InvalidCharacter, c);
				}
				while (c >= '0' && c <= '9')
				{
					b.Append(c);
					Read();
					c = Peek();
				}
			}
			if (c == 'e' || c == 'E')
			{
				b.Append('e');
				integer = false;
				Read();
				c = Read();
				if (c == '+')
				{
					c = Read();
				}
				else if (c == '-')
				{
					b.Append('-');
					c = Read();
				}
				if (c < '0' || c > '9')
				{
					throw Error(Errors.InvalidCharacter, c);
				}
				b.Append(c);
				c = Peek();
				while (c >= '0' && c <= '9')
				{
					b.Append(c);
					Read();
					c = Peek();
				}
			}
			string number = b.ToString();
			if (integer)
			{
				int val;
				if (int.TryParse(number, NumberStyles.Integer, CultureInfo.InvariantCulture, out val))
				{
					return new Token(new SalNumberLit(val, this.Location));
				}
				else
				{
					return new Token(new SalNumberLit(PPJ.Runtime.Sal.StrToNumber(number), this.Location));
				}
			}
			else
			{ // not integer
				double val;
				if (double.TryParse(number, NumberStyles.Float, CultureInfo.InvariantCulture, out val))
				{
					return new Token(new SalNumberLit((PPJ.Runtime.SalNumber)val, this.Location));
				}
				else
				{
					throw Error(Errors.InvalidNumberLiteral);
				}
			}
		}
	}
}
